/*
 * Copyright 2021 Thingstream AG
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdint.h>

#include <thingstream.h>

/* This currently requires the third-party library FlashStorage */
#include <FlashStorage.h>

#if !defined(ARDUINO_SODAQ_SARA) && !defined(ARDUINO_SODAQ_SFF)
#error This target is not supported
#endif

#include "tracker2.h"
#include "platform.h"
#include "nvmem.h"
#include "leds.h"
#include "timer.h"
#include "accel.h"
#include "modem_config.h"

static struct
{
    /* The writable copy of the flash config */
    NVMem *nvmem;

    /* We want to flash the LED every 15 seconds, so
     * we record the timestamp of the last flash,
     * and flash if it has changed. This is in units
     * of the 15-second RTC tick.
     * Also, we abuse it to cause a flash on motion.
     */
    unsigned int ledFlashTime;

    /* Set by interrupt handler on movement. */
    int activityDetected;

} state;


/* handler for accelerator interrupt */
static void activityDetector(void)
{
    ++state.activityDetected;
    state.ledFlashTime = ~0;  // force a flash now
}

/* Run one cycle of the tracker */
static void run_tracker(InvocationType mode)
{
    Tracker_run(mode,
                state.nvmem,
                0);

    // allow program to complete, then ensure LEDs have stopped
    // using the timer
    Led_Wait(true);
    Led_setRGB(0,0,0);

    // the config may have changed
    Accel_setSensitivity(state.nvmem->motionSensitivity);

    /* If we are moving right now, record it so that we trigger
     * a publish at 'interval.min'. Otherwise clear it, so that
     * we need to wait for fresh activity.
     */
    state.activityDetected = Accel_isActive();

    // configure 15-second timer ticks
    Timer_start(15*1024);
    Timer_triggerCount = 0;

    // and set the leds back to digital mode
    pinMode(LED_RED,   OUTPUT); digitalWrite(LED_RED,   HIGH);
    pinMode(LED_GREEN, OUTPUT); digitalWrite(LED_GREEN, HIGH);
    pinMode(LED_BLUE,  OUTPUT); digitalWrite(LED_BLUE,  HIGH);
}


void setup()
{
    Platform_init();
    Led_Init();
    Timer_init();
    Accel_init();

    state.nvmem = NVMem_init();

    Accel_setSensitivity(state.nvmem->motionSensitivity);

    /* A transition from high to low indicates movement */
    attachInterrupt(digitalPinToInterrupt(ACCEL_ACTIVITY_PIN),
                    activityDetector, FALLING);

    Thingstream_Util_printf("Configuring modem...\n");
    Platform_GsmEnable();
    setup_modem();

    Thingstream_Util_printf("Tracker starting...\n");
    run_tracker(eventPower);
}


/* abuse NUM_EVENTS since it is a dummy invalid value within the enum */
#define eventNone NUM_EVENTS

void loop()
{
    // TODO I'd like the first run of tracker to happen in here
    // rather than in setup().  Can't really think of an elegant
    // way of doing it other than having a static bool first-time
    // flag or similar. Maybe making 'ev' static, and setting
    // it to eventNone at the end of every run.

    InvocationType ev = eventNone;
    NVMem *nvmem = state.nvmem;

    // Timer_triggerCount measures time since last run, in 15-second
    // intervals.

    if (nvmem->interval.max == 0)
    {
        /* never send an update */
    }
    else if (Timer_triggerCount >= nvmem->interval.max * 4)
    {
        ev = eventTimer;
        // TODO record interval.max as the event parameter
    }
    else if (state.activityDetected &&
             Timer_triggerCount >= nvmem->interval.min * 4)
    {
        ev = eventMotion;
        // TODO record interval.min as the event parameter
        // TODO ideally want to distinguish which of the
        // two parts of the test became true first : if
        // the activity happened before the min time, it's
        // it's a timer + motion event, else it's a motion event.
    }

    if (ev != eventNone)
    {
        Timer_stop();  // make it available for the LEDs
        run_tracker(ev);
    }
    else if (state.ledFlashTime != Timer_triggerCount)
    {
        // just give a brief flash of the LED to show
        // we are still alive
        int led = state.activityDetected ? LED_GREEN : LED_BLUE;
        digitalWrite(led, 0);
        delay(20);
        digitalWrite(led, 1);
        state.ledFlashTime = Timer_triggerCount;

        if (Platform_DeepSleepEnabled())
        {
            Platform_low_power_wfi();
        }
        else
        {
            // Flash once more to indicate we are not going into deep sleep
            delay(200);
            digitalWrite(led, 0);
            delay(20);
            digitalWrite(led, 1);
            Platform_wfi();
        }
    }
}
