/*
 * Copyright 2021 Thingstream AG
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


/* An implementaion of Gnss_get_location() which retrieves NMEA
 * sentences from the M8Q device over I2C,
 */

#include <Arduino.h>
#include <stdint.h>
#include <stddef.h>

#include <thingstream.h>
#include "gnss.h"
#include "gps.h"
#include "platform.h"


uint8_t Gnss_get_location(uint32_t maxtime, int32_t *lat, int32_t *lon)
{
    uint8_t gpsFixQuality = 0;

    /* Turn on GPS and initialise NMEA parser */
    void *gps_cookie = Gps_init(true);

    uint32_t now = Thingstream_Platform_getTimeMillis();
    uint32_t gpsTimeout = now + maxtime;

    /* Wait for a fix */

    for (;;)
    {
        if (TIME_COMPARE(now, >=, gpsTimeout))
        {
            DEBUGOUT("Failed to latch GPS location\n");
            break;
        }

        uint8_t buffer[250];  // keep this less than 256
        uint32_t count = Platform_GpsRecv(buffer, sizeof(buffer));

        if (count)
        {
            Gps_callback(gps_cookie, buffer, count);
        }

        gpsFixQuality = Gps_getQuality();
        if (gpsFixQuality)
        {
            DEBUGOUT("GPS location determined successfully (%d)\n", gpsFixQuality);
            *lat = Gps_getLatitude();
            *lon = Gps_getLongitude();
            break;
        }

        delay(250);

        now = Thingstream_Platform_getTimeMillis();
    }

    Gps_init(false);

    return gpsFixQuality;
}
